/* USER CODE BEGIN Header */
/**
 ******************************************************************************
 * @file           : main.c
 * @brief          : Main program body
 ******************************************************************************
 * @attention
 *
 * <h2><center>&copy; Copyright (c) 2021 STMicroelectronics.
 * All rights reserved.</center></h2>
 *
 * This software component is licensed by ST under BSD 3-Clause license,
 * the "License"; You may not use this file except in compliance with the
 * License. You may obtain a copy of the License at:
 *                        opensource.org/licenses/BSD-3-Clause
 *
 ******************************************************************************
 */
/* USER CODE END Header */
/* Includes ------------------------------------------------------------------*/
#include "main.h"
#include "usb_host.h"

/* Private includes ----------------------------------------------------------*/
/* USER CODE BEGIN Includes */
typedef
		struct	{
	uint8_t			address;	// アドレスNo.
	uint32_t		bit;		// ビット位置
}
KEYB_TABLE;

#include "usbh_hid.h"
#include "usb_keytable.h"
#include "usbh_hid_gamepad.h"
/* USER CODE END Includes */

/* Private typedef -----------------------------------------------------------*/
/* USER CODE BEGIN PTD */
typedef
		enum
		{
	MODE_MOUSE = 0,
	MODE_KEYBOARD,
	MODE_GAMEPAD,
		}
MODE_MSX;

enum
{
	BUTTON_UP = 0,
	BUTTON_DOWN,
	BUTTON_LEFT,
	BUTTON_RIGHT,
	BUTTON_A,
	BUTTON_B,
	BUTTON_C,
	BUTTON_X,
	BUTTON_Y,
	BUTTON_Z,
	BUTTON_S,
	BUTTON_M,
	BUTTON_MAX
};

enum
{
	GAMEPAD_MODE1 = 0,
	GAMEPAD_MODE2,
	GAMEPAD_MODE3,
	GAMEPAD_MODE4,
	GAMEPAD_MODEMAX
};

/* USER CODE END PTD */

/* Private define ------------------------------------------------------------*/
/* USER CODE BEGIN PD */
#define KEYB_CTRL	        (0xe0)
#define KEYB_SHIFT	        (0xe1)
#define KEYB_ALT	        (0xe2)
#define KEYB_GUI	        (0xe3)

#define	 KEYB_UP			(0x52)
#define	 KEYB_DOWN			(0x51)
#define	 KEYB_LEFT			(0x50)
#define	 KEYB_RIGHT			(0x4f)
#define	 KEYB_SPC			(0x2c)
#define	 KEYB_RET	     	(0x28)
#define KEYB_M				(0x10)
#define KEYB_N				(0x11)
#define KEYB_Y				(0x1c)

#define	 ADDRESS_MAX        (0x10)

#define HATSW_UP           (0)
#define HATSW_UPRIGHT      (1)
#define HATSW_RIGHT        (2)
#define HATSW_DOWNRIGHT    (3)
#define HATSW_DOWN         (4)
#define HATSW_DOWNLEFT     (5)
#define HATSW_LEFT         (6)
#define HATSW_UPLEFT       (7)

#define RAPID_FAST         (1)
#define RAPID_SLOW         (2)
#define RAPID_MAX          (30000)

#define	 ADDRESS_MAX		(0x10)

#define MOUSE_UP			(0xF8)
#define MOUSE_CENTER		(0x80)
#define MOUSE_DOWN			(0x08)
/* USER CODE END PD */

/* Private macro -------------------------------------------------------------*/
/* USER CODE BEGIN PM */

/* USER CODE END PM */

/* Private variables ---------------------------------------------------------*/
TIM_HandleTypeDef htim11;

/* USER CODE BEGIN PV */

// USB
static KEYB_TABLE		usb_keytable[USB_SIZE];
uint32_t				msx_data[ADDRESS_MAX] = {0};

// デフォルトデータ
uint32_t				msx_reset = 0x00000000;

static MODE_MSX			mode_msx = MODE_MOUSE;
static uint32_t 		mode_gamepad = GAMEPAD_MODE1;

uint32_t				*msx_data0  = &msx_data[MSX_ID_A0];
uint32_t				*msx_data1  = &msx_data[MSX_ID_A1];
uint32_t				*msx_data2  = &msx_data[MSX_ID_A2];
uint32_t				*msx_data3  = &msx_data[MSX_ID_A3];
uint32_t				*msx_data4  = &msx_data[MSX_ID_A4];
uint32_t				*msx_data5  = &msx_data[MSX_ID_A5];
uint32_t				*msx_data6  = &msx_data[MSX_ID_A6];
uint32_t				*msx_data7  = &msx_data[MSX_ID_A7];
uint32_t				*msx_data8  = &msx_data[MSX_ID_A8];
uint32_t				*msx_data9  = &msx_data[MSX_ID_A9];
uint32_t				*msx_data10 = &msx_data[MSX_ID_A10];

// GamePad
static uint8_t		pad_con_x     = 0;
static uint8_t		pad_con_y     = 0;
static uint8_t		pad_con_hatsw = 0;
static uint8_t		pad_bot_a     = 0;
static uint8_t		pad_bot_b     = 0;
static uint8_t		pad_bot_c     = 0;
static uint8_t		pad_bot_x     = 0;
static uint8_t		pad_bot_y     = 0;
static uint8_t		pad_bot_z     = 0;
static uint8_t		pad_bot_s     = 0;
static uint8_t		pad_bot_m     = 0;

static uint32_t		rapid_count = 0;
static uint32_t		rapid_speed = RAPID_FAST;

const static uint8_t  gamepad_map[GAMEPAD_MODEMAX][BUTTON_MAX] =
{
		//    [Up]   [Down]     [Left]     [Right]        [A]         [B]       [C]         [START]      [MODE]
		{ KEYB_UP,	KEYB_DOWN,	KEYB_LEFT,	KEYB_RIGHT,	KEYB_SPC,	KEYB_M,		KEYB_N,		KEYB_RET,	KEYB_Y },
		{ KEYB_UP,	KEYB_DOWN,	KEYB_LEFT,	KEYB_RIGHT,	KEYB_N,		KEYB_M,		KEYB_SPC,	KEYB_RET,	KEYB_Y },
		{ KEYB_UP,	KEYB_DOWN,	KEYB_LEFT,	KEYB_RIGHT,	KEYB_SPC,	KEYB_SHIFT,	KEYB_SPC,	KEYB_RET,	KEYB_Y },
		{ KEYB_UP,	KEYB_DOWN,	KEYB_LEFT,	KEYB_RIGHT,	KEYB_SHIFT,	KEYB_SPC,	KEYB_SHIFT,	KEYB_RET,	KEYB_Y },
};
/* USER CODE END PV */

/* Private function prototypes -----------------------------------------------*/
void SystemClock_Config(void);
static void MX_GPIO_Init(void);
static void MX_TIM11_Init(void);
void MX_USB_HOST_Process(void);

/* USER CODE BEGIN PFP */
static void set_usb_mouse(USBH_HandleTypeDef *phost);
static void set_usb_keycode(USBH_HandleTypeDef *phost);
static void set_usb_gamepad(USBH_HandleTypeDef *phost);
static void msx_gamepad(void);
/* USER CODE END PFP */

/* Private user code ---------------------------------------------------------*/
/* USER CODE BEGIN 0 */

/* USER CODE END 0 */

/**
  * @brief  The application entry point.
  * @retval int
  */
int main(void)
{
  /* USER CODE BEGIN 1 */

  /* USER CODE END 1 */

  /* MCU Configuration--------------------------------------------------------*/

  /* Reset of all peripherals, Initializes the Flash interface and the Systick. */
  HAL_Init();

  /* USER CODE BEGIN Init */

  /* USER CODE END Init */

  /* Configure the system clock */
  SystemClock_Config();

  /* USER CODE BEGIN SysInit */

  /* USER CODE END SysInit */

  /* Initialize all configured peripherals */
  MX_GPIO_Init();
  MX_USB_HOST_Init();
  MX_TIM11_Init();
  /* USER CODE BEGIN 2 */
	// VBUS無効化
	USB_OTG_FS->GCCFG |= USB_OTG_GCCFG_NOVBUSSENS;

	// デフォルトデータセット
	msx_reset = (uint32_t)(MSX_D0_Pin|MSX_D1_Pin|MSX_D2_Pin|MSX_D3_Pin|MSX_D4_Pin|MSX_D5_Pin|MSX_D6_Pin|MSX_D7_Pin);

	// ビットシフト時間短縮のため事前に並べ替えておく
	for( uint32_t index = 0; index < USB_SIZE; index++ ){
		usb_keytable[index].address = msx_usb_keytable[index].address;
		usb_keytable[index].bit = 0x00000000;
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x01 ? ((uint32_t)MSX_D0_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x02 ? ((uint32_t)MSX_D1_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x04 ? ((uint32_t)MSX_D2_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x08 ? ((uint32_t)MSX_D3_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x10 ? ((uint32_t)MSX_D4_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x20 ? ((uint32_t)MSX_D5_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x40 ? ((uint32_t)MSX_D6_Pin << 16u) : 0x00000000);
		usb_keytable[index].bit |= (msx_usb_keytable[index].bit & 0x80 ? ((uint32_t)MSX_D7_Pin << 16u) : 0x00000000);
	}

	HAL_Delay(50);

	// タイマ開始
	HAL_TIM_Base_Start_IT(&htim11);
  /* USER CODE END 2 */

  /* Infinite loop */
  /* USER CODE BEGIN WHILE */
  while (1)
  {
    /* USER CODE END WHILE */
    MX_USB_HOST_Process();

    /* USER CODE BEGIN 3 */
  }
  /* USER CODE END 3 */
}

/**
  * @brief System Clock Configuration
  * @retval None
  */
void SystemClock_Config(void)
{
  RCC_OscInitTypeDef RCC_OscInitStruct = {0};
  RCC_ClkInitTypeDef RCC_ClkInitStruct = {0};

  /** Configure the main internal regulator output voltage
  */
  __HAL_RCC_PWR_CLK_ENABLE();
  __HAL_PWR_VOLTAGESCALING_CONFIG(PWR_REGULATOR_VOLTAGE_SCALE2);

  /** Initializes the RCC Oscillators according to the specified parameters
  * in the RCC_OscInitTypeDef structure.
  */
  RCC_OscInitStruct.OscillatorType = RCC_OSCILLATORTYPE_HSE;
  RCC_OscInitStruct.HSEState = RCC_HSE_ON;
  RCC_OscInitStruct.PLL.PLLState = RCC_PLL_ON;
  RCC_OscInitStruct.PLL.PLLSource = RCC_PLLSOURCE_HSE;
  RCC_OscInitStruct.PLL.PLLM = 25;
  RCC_OscInitStruct.PLL.PLLN = 336;
  RCC_OscInitStruct.PLL.PLLP = RCC_PLLP_DIV4;
  RCC_OscInitStruct.PLL.PLLQ = 7;
  if (HAL_RCC_OscConfig(&RCC_OscInitStruct) != HAL_OK)
  {
    Error_Handler();
  }

  /** Initializes the CPU, AHB and APB buses clocks
  */
  RCC_ClkInitStruct.ClockType = RCC_CLOCKTYPE_HCLK|RCC_CLOCKTYPE_SYSCLK
                              |RCC_CLOCKTYPE_PCLK1|RCC_CLOCKTYPE_PCLK2;
  RCC_ClkInitStruct.SYSCLKSource = RCC_SYSCLKSOURCE_PLLCLK;
  RCC_ClkInitStruct.AHBCLKDivider = RCC_SYSCLK_DIV1;
  RCC_ClkInitStruct.APB1CLKDivider = RCC_HCLK_DIV2;
  RCC_ClkInitStruct.APB2CLKDivider = RCC_HCLK_DIV1;

  if (HAL_RCC_ClockConfig(&RCC_ClkInitStruct, FLASH_LATENCY_2) != HAL_OK)
  {
    Error_Handler();
  }
}

/**
  * @brief TIM11 Initialization Function
  * @param None
  * @retval None
  */
static void MX_TIM11_Init(void)
{

  /* USER CODE BEGIN TIM11_Init 0 */

  /* USER CODE END TIM11_Init 0 */

  /* USER CODE BEGIN TIM11_Init 1 */

  /* USER CODE END TIM11_Init 1 */
  htim11.Instance = TIM11;
  htim11.Init.Prescaler = 83;
  htim11.Init.CounterMode = TIM_COUNTERMODE_UP;
  htim11.Init.Period = 33333;
  htim11.Init.ClockDivision = TIM_CLOCKDIVISION_DIV1;
  htim11.Init.AutoReloadPreload = TIM_AUTORELOAD_PRELOAD_ENABLE;
  if (HAL_TIM_Base_Init(&htim11) != HAL_OK)
  {
    Error_Handler();
  }
  /* USER CODE BEGIN TIM11_Init 2 */

  /* USER CODE END TIM11_Init 2 */

}

/**
  * @brief GPIO Initialization Function
  * @param None
  * @retval None
  */
static void MX_GPIO_Init(void)
{
  GPIO_InitTypeDef GPIO_InitStruct = {0};

  /* GPIO Ports Clock Enable */
  __HAL_RCC_GPIOH_CLK_ENABLE();
  __HAL_RCC_GPIOA_CLK_ENABLE();
  __HAL_RCC_GPIOB_CLK_ENABLE();

  /*Configure GPIO pin Output Level */
  HAL_GPIO_WritePin(GPIOA, MSX_D0_Pin|MSX_D1_Pin|MSX_D2_Pin|MSX_D3_Pin
                          |MSX_D4_Pin|MSX_D5_Pin|MSX_D6_Pin|MSX_D7_Pin, GPIO_PIN_SET);

  /*Configure GPIO pins : MSX_D0_Pin MSX_D1_Pin MSX_D2_Pin MSX_D3_Pin
                           MSX_D4_Pin MSX_D5_Pin MSX_D6_Pin MSX_D7_Pin */
  GPIO_InitStruct.Pin = MSX_D0_Pin|MSX_D1_Pin|MSX_D2_Pin|MSX_D3_Pin
                          |MSX_D4_Pin|MSX_D5_Pin|MSX_D6_Pin|MSX_D7_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_OUTPUT_OD;
  GPIO_InitStruct.Pull = GPIO_NOPULL;
  GPIO_InitStruct.Speed = GPIO_SPEED_FREQ_VERY_HIGH;
  HAL_GPIO_Init(GPIOA, &GPIO_InitStruct);

  /*Configure GPIO pins : MSX_A9_Pin MSX_A10_Pin MSX_A4_Pin MSX_A5_Pin
                           MSX_A6_Pin MSX_A7_Pin MSX_A8_Pin MSX_A0_Pin
                           MSX_A1_Pin MSX_A2_Pin MSX_A3_Pin */
  GPIO_InitStruct.Pin = MSX_A9_Pin|MSX_A10_Pin|MSX_A4_Pin|MSX_A5_Pin
                          |MSX_A6_Pin|MSX_A7_Pin|MSX_A8_Pin|MSX_A0_Pin
                          |MSX_A1_Pin|MSX_A2_Pin|MSX_A3_Pin;
  GPIO_InitStruct.Mode = GPIO_MODE_IT_FALLING;
  GPIO_InitStruct.Pull = GPIO_PULLUP;
  HAL_GPIO_Init(GPIOB, &GPIO_InitStruct);

  /* EXTI interrupt init*/
  HAL_NVIC_SetPriority(EXTI0_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI0_IRQn);

  HAL_NVIC_SetPriority(EXTI1_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI1_IRQn);

  HAL_NVIC_SetPriority(EXTI9_5_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI9_5_IRQn);

  HAL_NVIC_SetPriority(EXTI15_10_IRQn, 0, 0);
  HAL_NVIC_EnableIRQ(EXTI15_10_IRQn);

}

/* USER CODE BEGIN 4 */
/**
 * @brief  Period elapsed callback in non-blocking mode
 * @param  htim TIM handle
 * @retval None
 */
void HAL_TIM_PeriodElapsedCallback(TIM_HandleTypeDef *htim)
{
	/* Prevent unused argument(s) compilation warning */
	//  UNUSED(htim);

	/* NOTE : This function should not be modified, when the callback is needed,
            the HAL_TIM_PeriodElapsedCallback could be implemented in the user file
	 */
	if (htim == &htim11){
		msx_gamepad();
	}
}

/**
 * @brief  The function is a callback about HID Data events
 *  @param  phost: Selected device
 * @retval None
 */
void USBH_HID_EventCallback(USBH_HandleTypeDef *phost)
{
	/* Prevent unused argument(s) compilation warning */
	//  UNUSED(phost);
	HID_TypeTypeDef			hid_type	= USBH_HID_GetDeviceType(phost);

	// Get HID Mouse Information
	if(hid_type == HID_MOUSE){
		mode_msx = MODE_MOUSE;
		set_usb_mouse(phost);
	}
	// Get HID Keyboard Information
	else if(hid_type == HID_KEYBOARD){
		mode_msx = MODE_KEYBOARD;
		set_usb_keycode(phost);
	}
	// Get HID Gamepad Information
	else{
		mode_msx = MODE_GAMEPAD;
		set_usb_gamepad(phost);
	}
}

static void set_usb_mouse(USBH_HandleTypeDef *phost)
{
	HID_MOUSE_Info_TypeDef	*mouse_info = USBH_HID_GetMouseInfo(phost);
	uint8_t mouse_x = mouse_info->x;
	uint8_t mouse_y = mouse_info->y;
	uint8_t mouse_l = mouse_info->buttons[0];
	uint8_t mouse_r = mouse_info->buttons[1];

	uint16_t button_up    = KEYB_UP;
	uint16_t button_down  = KEYB_DOWN;
	uint16_t button_left  = KEYB_LEFT;
	uint16_t button_right = KEYB_RIGHT;

	uint16_t button_a     = KEYB_SPC;
	uint16_t button_b     = KEYB_SHIFT;

	if( mouse_info == NULL ){
		return;
	}

	memset( msx_data, 0, sizeof(msx_data) );

	// A
	if( mouse_l ){
		msx_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
	}
	// B
	if( mouse_r ){
		msx_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
	}

	// UP
	if( mouse_y < MOUSE_UP && mouse_y > MOUSE_CENTER ){
		msx_data[usb_keytable[button_up].address] |=  (uint32_t)usb_keytable[button_up].bit;
	}
	// DOWN
	else if( mouse_y > MOUSE_DOWN && mouse_y < MOUSE_CENTER ){
		msx_data[usb_keytable[button_down].address] |=  (uint32_t)usb_keytable[button_down].bit;
	}
	// LEFT
	if( mouse_x < MOUSE_UP && mouse_x > MOUSE_CENTER ){
		msx_data[usb_keytable[button_left].address] |=  (uint32_t)usb_keytable[button_left].bit;
	}
	// RIGHT
	else if( mouse_x > MOUSE_DOWN  && mouse_x < MOUSE_CENTER ){
		msx_data[usb_keytable[button_right].address] |=  (uint32_t)usb_keytable[button_right].bit;
	}
}

static void set_usb_keycode(USBH_HandleTypeDef *phost)
{
	HID_KEYBD_Info_TypeDef	*keyb_info = USBH_HID_GetKeybdInfo(phost);

	if( keyb_info == NULL ){
		return;
	}

	memset( msx_data, 0, sizeof(msx_data) );

	// shift
	if( keyb_info->lshift || keyb_info->rshift ){
		msx_data[usb_keytable[KEYB_SHIFT].address] |= ((uint32_t)usb_keytable[KEYB_SHIFT].bit);
	}
	// ctrl
	if( keyb_info->lctrl || keyb_info->rctrl ){
		msx_data[usb_keytable[KEYB_CTRL].address] |= ((uint32_t)usb_keytable[KEYB_CTRL].bit);
	}
	// alt - Grapth
	if( keyb_info->lalt || keyb_info->ralt ){
		msx_data[usb_keytable[KEYB_ALT].address]  |= ((uint32_t)usb_keytable[KEYB_ALT].bit);
	}
	// gui - Kana
	if( keyb_info->lgui || keyb_info->rgui ){
		msx_data[usb_keytable[KEYB_GUI].address]  |= ((uint32_t)usb_keytable[KEYB_GUI].bit);
	}

	// USBキーボードの押されているキー
	for( uint32_t loop = 0; loop < 6; loop++ ){
		uint8_t		usb_code = keyb_info->keys[loop];
		if( usb_code != 0 ){
			msx_data[usb_keytable[usb_code].address] |= ((uint32_t)usb_keytable[usb_code].bit);
		}
	}
}

static void set_usb_gamepad(USBH_HandleTypeDef *phost)
{
	HID_GAMEPAD_Info_TypeDef	*gamepad_info = USBH_HID_GetGamepadInfo(phost);

	if( gamepad_info == NULL ){
		return;
	}

	pad_con_x     = gamepad_info->x;
	pad_con_y     = gamepad_info->y;
	pad_con_hatsw = gamepad_info->hatsw;
	pad_bot_a     = gamepad_info->button[5];
	pad_bot_b     = gamepad_info->button[1];
	pad_bot_x     = gamepad_info->button[4];
	pad_bot_y     = gamepad_info->button[0];
	pad_bot_c     = gamepad_info->button[2];
	pad_bot_z     = gamepad_info->button[3];
	pad_bot_s     = gamepad_info->button[9];
	pad_bot_m     = gamepad_info->button[8];
}

static void msx_gamepad(void)
{
	uint16_t button_up    = gamepad_map[mode_gamepad][BUTTON_UP];
	uint16_t button_down  = gamepad_map[mode_gamepad][BUTTON_DOWN];
	uint16_t button_left  = gamepad_map[mode_gamepad][BUTTON_LEFT];
	uint16_t button_right = gamepad_map[mode_gamepad][BUTTON_RIGHT];
	uint16_t button_a     = gamepad_map[mode_gamepad][BUTTON_A];
	uint16_t button_b     = gamepad_map[mode_gamepad][BUTTON_B];
	uint16_t button_c     = gamepad_map[mode_gamepad][BUTTON_C];
	uint16_t button_s     = gamepad_map[mode_gamepad][BUTTON_S];
	uint16_t button_m     = gamepad_map[mode_gamepad][BUTTON_M];

	if( mode_msx != MODE_GAMEPAD ){
		return;
	}

	rapid_count++;
	if( rapid_count > RAPID_MAX )		rapid_count = 0;

	memset( msx_data, 0, sizeof(msx_data) );

	if( pad_bot_a ){
		msx_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
	}
	else if( pad_bot_x && (rapid_count & rapid_speed)){
		msx_data[usb_keytable[button_a].address] |=  (uint32_t)usb_keytable[button_a].bit;
	}
	if( pad_bot_b ){
		msx_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
	}
	else if( pad_bot_y && (rapid_count & rapid_speed)){
		msx_data[usb_keytable[button_b].address] |=  (uint32_t)usb_keytable[button_b].bit;
	}
	if( pad_bot_c ){
		msx_data[usb_keytable[button_c].address] |=  (uint32_t)usb_keytable[button_c].bit;
	}
	else if( pad_bot_z && (rapid_count & rapid_speed)){
		msx_data[usb_keytable[button_c].address] |=  (uint32_t)usb_keytable[button_c].bit;
	}

	// UP
	if( pad_con_y < 0x40 || pad_con_hatsw == HATSW_UP || pad_con_hatsw == HATSW_UPLEFT || pad_con_hatsw == HATSW_UPRIGHT ){
		msx_data[usb_keytable[button_up].address] |=  (uint32_t)usb_keytable[button_up].bit;
	}
	// DOWN
	else if( pad_con_y > 0xc0 || pad_con_hatsw == HATSW_DOWN || pad_con_hatsw == HATSW_DOWNLEFT || pad_con_hatsw == HATSW_DOWNRIGHT ){
		msx_data[usb_keytable[button_down].address] |=  (uint32_t)usb_keytable[button_down].bit;
	}
	// LEFT
	if( pad_con_x < 0x40 || pad_con_hatsw == HATSW_LEFT || pad_con_hatsw == HATSW_UPLEFT || pad_con_hatsw == HATSW_DOWNLEFT ){
		msx_data[usb_keytable[button_left].address] |=  (uint32_t)usb_keytable[button_left].bit;
	}
	// RIGHT
	else if( pad_con_x > 0xc0  || pad_con_hatsw == HATSW_RIGHT || pad_con_hatsw == HATSW_UPRIGHT || pad_con_hatsw == HATSW_DOWNRIGHT ){
		msx_data[usb_keytable[button_right].address] |=  (uint32_t)usb_keytable[button_right].bit;
	}

	if( pad_bot_s ){	// START
		msx_data[usb_keytable[button_s].address] |=  (uint32_t)usb_keytable[button_s].bit;
	}

	if( pad_bot_m ){	// MODE
		msx_data[usb_keytable[button_m].address] |=  (uint32_t)usb_keytable[button_m].bit;
	}

	// Change GAMEPAD MODE
	// START
	if( pad_bot_s ){
		// LEFT
		if( pad_con_x < 0x40 || pad_con_hatsw == HATSW_LEFT || pad_con_hatsw == HATSW_UPLEFT || pad_con_hatsw == HATSW_DOWNLEFT ){
			mode_gamepad = GAMEPAD_MODE1;
		}
		// UP
		else if( pad_con_y < 0x40 || pad_con_hatsw == HATSW_UP || pad_con_hatsw == HATSW_UPLEFT || pad_con_hatsw == HATSW_UPRIGHT ){
			mode_gamepad = GAMEPAD_MODE2;
		}
		// RIGHT
		else if( pad_con_x > 0xc0  || pad_con_hatsw == HATSW_RIGHT || pad_con_hatsw == HATSW_UPRIGHT || pad_con_hatsw == HATSW_DOWNRIGHT ){
			mode_gamepad = GAMEPAD_MODE3;
		}
		// DOWN
		else if( pad_con_y > 0xc0 || pad_con_hatsw == HATSW_DOWN || pad_con_hatsw == HATSW_DOWNLEFT || pad_con_hatsw == HATSW_DOWNRIGHT ){
			mode_gamepad = GAMEPAD_MODE4;
		}
		else if( pad_bot_a || pad_bot_x ){
			rapid_speed = RAPID_SLOW;
		}
		else if( pad_bot_c || pad_bot_z ){
			rapid_speed = RAPID_FAST;
		}
	}
}

/**
 * @brief  EXTI line detection callbacks.
 * @param  GPIO_Pin: Specifies the pins connected EXTI line
 * @retval None
 */
void HAL_GPIO_EXTI_Callback(uint16_t GPIO_Pin)
{
	/* Prevent unused argument(s) compilation warning */
	//UNUSED(GPIO_Pin);
	/* NOTE: This function Should not be modified, when the callback is needed,
           the HAL_GPIO_EXTI_Callback could be implemented in the user file
	 */
}
/* USER CODE END 4 */

/**
  * @brief  This function is executed in case of error occurrence.
  * @retval None
  */
void Error_Handler(void)
{
  /* USER CODE BEGIN Error_Handler_Debug */
  /* User can add his own implementation to report the HAL error return state */
  __disable_irq();
  while (1)
  {
  }
  /* USER CODE END Error_Handler_Debug */
}

#ifdef  USE_FULL_ASSERT
/**
  * @brief  Reports the name of the source file and the source line number
  *         where the assert_param error has occurred.
  * @param  file: pointer to the source file name
  * @param  line: assert_param error line source number
  * @retval None
  */
void assert_failed(uint8_t *file, uint32_t line)
{
  /* USER CODE BEGIN 6 */
  /* User can add his own implementation to report the file name and line number,
     ex: printf("Wrong parameters value: file %s on line %d\r\n", file, line) */
  /* USER CODE END 6 */
}
#endif /* USE_FULL_ASSERT */
